/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.valueprovider;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.jface.viewers.ColumnViewer;
import org.eclipse.jface.viewers.EditingSupport;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IDerivedAnnotation;

/**
 * Base class for {@link IAnnotationValueProvider} for derived annotations (i.e., based on
 * {@link IDerivedAnnotation}).
 * 
 * @author barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public abstract class DerivedAnnotationValueProviderBase<T extends IDerivedAnnotation<?>> extends
		EStructuralFeatureValueProviderBase<T> {

	/** Key representing the derived / calculated value. */
	private String derivedKey;

	/** List of all supported instance keys. */
	private List<String> keys = new ArrayList<String>();

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a pure derived annotations (i.e., an
	 * {@link IDerivedAnnotation} that does not contain any additional {@link EStructuralFeature}s.
	 */
	public DerivedAnnotationValueProviderBase(EClass annotatedSpecificationEClass) {
		super(annotatedSpecificationEClass);
		handlesMultipleEstructuralFeatures = false;
		derivedKey = DEFAULT_KEY;

		keys.add(DEFAULT_KEY);
	}

	/**
	 * Constructs a {@link IAnnotationValueProvider} for a derived annotations (i.e., based on an
	 * {@link IDerivedAnnotation}) contains additional {@link EStructuralFeature}s.
	 * 
	 * @param annotatedSpecificationEClass
	 *            {@link EClass} implementing the annotation
	 * @param structuralFeatureDescriptorMap
	 *            map: instance key -> {@link EStructuralFeatureDescriptor}s
	 * @param derivedKey
	 *            key representing the derived / calculated value (must not be contained in
	 *            structuralFeatureMap).
	 */
	public DerivedAnnotationValueProviderBase(EClass annotatedSpecificationEClass,
			Map<String, EStructuralFeatureDescriptor> structuralFeatureDescriptorMap,
			String derivedKey) {
		super(annotatedSpecificationEClass, structuralFeatureDescriptorMap);

		this.derivedKey = derivedKey;
		keys.addAll(structuralFeatureDescriptorMap.keySet());
		keys.add(derivedKey);

	}

	/**
	 * Predicate if the given instanceKey is the key representing the derived / calculated value
	 * provided by this {@link IAnnotationValueProvider}.
	 */
	private boolean isDerivedKey(String instanceKey) {
		return (instanceKey == DEFAULT_KEY) || (instanceKey.equals(derivedKey));
	}

	/** {@inheritDoc} */
	@Override
	public boolean canEdit(T specification, String instanceKey) {
		if(isDerivedKey(instanceKey)) {
			return false;
		}
		return super.canEdit(specification, instanceKey);
	}

	/** {@inheritDoc} */
	@SuppressWarnings("unchecked")
	@Override
	public <U> U getAnnotationValue(T specification, String instanceKey) throws Exception {
		if(isDerivedKey(instanceKey)) {
			return (U)specification.getValue();
		}

		return super.getAnnotationValue(specification, instanceKey);
	}

	/** {@inheritDoc} */
	@Override
	public EditingSupport createEditingSupport(ColumnViewer viewer,
			Class<? extends IAnnotatedSpecification> clazz, String instanceKey) throws Exception {

		if(isDerivedKey(instanceKey)) {
			return null;
		}

		return super.createEditingSupport(viewer, clazz, instanceKey);
	}

	/** {@inheritDoc} */
	@Override
	public List<String> getInstanceKeys(T specification) {
		return keys;
	}

}
