/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view;

import org.eclipse.emf.common.notify.Notification;
import org.eclipse.emf.common.notify.impl.NotificationImpl;
import org.eclipse.emf.common.util.EMap;
import org.eclipse.emf.ecore.EClass;
import org.eclipse.emf.ecore.EStructuralFeature;
import org.eclipse.jface.dialogs.IInputValidator;
import org.eclipse.jface.dialogs.InputDialog;
import org.eclipse.jface.dialogs.MessageDialog;
import org.eclipse.jface.viewers.ColumnLabelProvider;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.ViewerCell;
import org.eclipse.jface.viewers.ViewerColumn;
import org.eclipse.jface.window.Window;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.TableEditor;
import org.eclipse.swt.events.DisposeEvent;
import org.eclipse.swt.events.DisposeListener;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Listener;
import org.eclipse.swt.widgets.Shell;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.TableItem;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.IAnnotationValueService;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * This class is responsible to construct a column which holds a button that allows adding new
 * instance for multi-instance annotations.
 * 
 * This class is in part based on code from the class {@link TableViewerColumn}.
 * 
 * @author diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED
 */
public class CreateAnnotationInstanceColumn extends ViewerColumn {

	/** the column to be added to the table */
	private TableColumn column;

	/** annotation specification this column refers to */
	private IAnnotatedSpecification columnSpec;

	/** annotation entry the specification {@code columnSpec} is associated with */
	private AnnotationEntry columnSpecAnnEntry;

	/** the button which opens a dialog to create new values for the specification {@code columnSpec} */
	private Button crButton;

	/**
	 * Creates a new viewer column for the given {@link TableViewer} on a new {@link TableColumn}
	 * with the given style bits. The column is added at the
	 * end of the list of columns.
	 * 
	 * @param viewer
	 *            the table viewer to which this column belongs
	 * @param style
	 *            the style used to create the column, for applicable style bits
	 *            see {@link TableColumn}
	 * @see TableColumn#TableColumn(Table, int)
	 */
	public CreateAnnotationInstanceColumn(TableViewer viewer, int style) {
		this(viewer, style, -1);
	}

	/**
	 * Creates a new viewer column for the given {@link TableViewer} on a new {@link TableColumn}
	 * with the given style bits. The column is inserted at
	 * the given index into the list of columns.
	 * 
	 * @param viewer
	 *            the table viewer to which this column belongs
	 * @param style
	 *            the style used to create the column, for applicable style bits
	 *            see {@link TableColumn}
	 * @param index
	 *            the index at which to place the newly created column
	 * @see TableColumn#TableColumn(Table, int, int)
	 */
	public CreateAnnotationInstanceColumn(TableViewer viewer, int style, int index) {
		this(viewer, createColumn(viewer.getTable(), style, index));
	}

	/**
	 * This class creates a standard table columns by calling the default of this class which is
	 * taken form {@link TableViewerColumn}. Additionally, it sets some specific values that are
	 * associated with the given annotation specification {@code spec}. Furthermore, it manages the
	 * creation and the functionality of the 'create' button and the associated dialog.
	 * 
	 * @param viewer
	 *            the table in which the column shall appear
	 * @param style
	 *            the column's style
	 * @param spec
	 *            the annotation specification that this class shall refer to
	 * @param entry
	 *            the AnnotationEntry associated with the given specification {@code spec}
	 */
	public CreateAnnotationInstanceColumn(TableViewer viewer, int style,
			IAnnotatedSpecification spec, AnnotationEntry entry) {
		this(viewer, style, -1);

		columnSpec = spec;
		columnSpecAnnEntry = entry;
		column.setText(entry.getSpecificationAnnotationName(spec.getClass()));

		setLabelProvider(new CreateNewAnnotationColumn());
		column.addDisposeListener(new DisposeListener() {

			@Override
			public void widgetDisposed(DisposeEvent e) {
				if(crButton != null) {
					crButton.dispose();
					crButton = null;
				}
			}
		});
	}

	/**
	 * Creates a new viewer column for the given {@link TableViewer} on the given
	 * {@link TableColumn}.
	 * 
	 * @param viewer
	 *            the table viewer to which this column belongs
	 * @param column
	 *            the underlying table column
	 */
	public CreateAnnotationInstanceColumn(TableViewer viewer, TableColumn column) {
		super(viewer, column);
		this.column = column;
	}

	/** creates the column by instantiating TableColumn */
	private static TableColumn createColumn(Table table, int style, int index) {
		if(index >= 0) {
			return new TableColumn(table, style, index);
		}

		return new TableColumn(table, style);
	}

	/**
	 * @return the underlying SWT table column
	 */
	public TableColumn getColumn() {
		return column;
	}

	/**
	 * creates the content of a new column via the update method of the table viewer to provide a
	 * "Create" button in the first row such that new values can be added to the table
	 */
	private class CreateNewAnnotationColumn extends ColumnLabelProvider {
		/** {@inheritDoc} */
		@Override
		public void update(ViewerCell cell) {

			TableItem item = (TableItem)cell.getItem();

			if(columnSpec != null && crButton == null) {
				crButton = new Button((Composite)cell.getViewerRow().getControl(), SWT.NONE);
				crButton.setText("Create");

				crButton.addListener(SWT.Selection, new CreateButtonListener());

				TableEditor editor = new TableEditor(item.getParent());
				editor.grabHorizontal = true;
				editor.grabVertical = true;
				editor.setEditor(crButton, item, cell.getColumnIndex());
				editor.layout();
			}
		}
	}

	/**
	 * This Listener watches the "Create" button such that a dialog is opened if pressed and a new
	 * key can be added to the specification type of this column
	 */
	private class CreateButtonListener implements Listener {
		/** Buffer holding the annotation's map to support the get/set functions of the model */

		/** {@inheritDoc} */
		@Override
		public void handleEvent(Event event) {
			if(event.widget == crButton) {
				InputDialog dlg =
						new InputDialog(Display.getCurrent().getActiveShell(), "Create Annotation",
								"Enter the name of the new annotation", "",
								new AnnotationNameValidator());
				if(dlg.open() == Window.OK) {
					final String instanceKey = dlg.getValue();

					// Register the new annotation instance with the model, by writing a null value
					// for the input key provided by the user. A write transaction / a switch to the
					// model context is required to modify the from the GUI context.
					ITopLevelElement modelContext =
							IPersistencyService.INSTANCE.getTopLevelElementFor(columnSpecAnnEntry
									.getModelElement());
					modelContext.runAsCommand(new Runnable() {

						@Override
						public void run() {
							try {
								// Populate new instance (in order to make the corresponding cells
								// editable
								String value = "";
								if(!columnSpec.eClass().getEStructuralFeatures().isEmpty()) {
									EStructuralFeature dynamicInstanceFeature =
											columnSpec.eClass().getEStructuralFeatures().get(0);

									// Annotations that can be dynamically instantiated are
									// implemented using an appropriate EMap.
									if(dynamicInstanceFeature.getEType().getInstanceTypeName()
											.equals("java.util.Map$Entry")) {

										// Determine value type
										EStructuralFeature valueFeature =
												((EClass)dynamicInstanceFeature.getEType())
														.getEStructuralFeature("value");
										if(Number.class.isAssignableFrom(valueFeature.getEType()
												.getInstanceClass())) {

											// Initialize numbers with 0
											// (initialization fails with "")
											value = "0";
										}
									}

								}

								for(AnnotationEntry entry : IAnnotationValueService.INSTANCE
										.getValues(columnSpecAnnEntry.getModelElement())) {

									entry.setSpecificationValue(value, columnSpec.getClass(),
											instanceKey);
								}

								// Inform others, e.g. the respective annotation view that a new
								// instance of this annotation has been created.
								columnSpecAnnEntry.getModelElement().eNotify(
										new NotificationImpl(Notification.SET, null, null) {
											@Override
											public Object getNotifier() {
												return columnSpecAnnEntry.getModelElement();
											}
										});
							} catch(IllegalArgumentException e) {
								// e.printStackTrace();
								MessageDialog
										.openError(
												new Shell(),
												"ERROR",
												"The value you entered does not have the desired type.\nDetailed message:\n" +
														e.getMessage());
							} catch(Exception e) {
								MessageDialog.openError(new Shell(), "ERROR", e.getMessage());
							}
						}
					});
				}
			}
		}
	}

	/** Provides a validator which checks if an input string is not an existing annotation name */
	class AnnotationNameValidator implements IInputValidator {
		/**
		 * Validates the String. Returns null for no error, or an error message
		 * 
		 * @param inputText
		 *            the String to validate
		 * @return String
		 */
		@Override
		public String isValid(String inputText) {
			@SuppressWarnings("unchecked") EMap<String, ?> kVMap =
					(EMap<String, ?>)columnSpecAnnEntry
							.getSpecificationValue(columnSpec.getClass());
			if(kVMap != null) {
				if(kVMap.containsKey(inputText))
					return "You cannot create a new key that already exists for the selected specification class.";
			}
			return null;
		}
	}
}
