/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.view;

import static org.fortiss.tooling.kernel.utils.EcoreUtils.pickInstanceOf;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import java.util.TreeSet;

import org.eclipse.emf.ecore.EObject;
import org.eclipse.jface.viewers.ArrayContentProvider;
import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.jface.viewers.TableViewerColumn;
import org.eclipse.jface.viewers.Viewer;
import org.eclipse.jface.viewers.ViewerComparator;
import org.eclipse.swt.SWT;
import org.eclipse.swt.custom.CCombo;
import org.eclipse.swt.events.ModifyEvent;
import org.eclipse.swt.events.ModifyListener;
import org.eclipse.swt.events.SelectionAdapter;
import org.eclipse.swt.events.SelectionEvent;
import org.eclipse.swt.layout.FillLayout;
import org.eclipse.swt.layout.GridData;
import org.eclipse.swt.layout.GridLayout;
import org.eclipse.swt.layout.RowLayout;
import org.eclipse.swt.widgets.Button;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.swt.widgets.Display;
import org.eclipse.swt.widgets.Event;
import org.eclipse.swt.widgets.Label;
import org.eclipse.swt.widgets.Table;
import org.eclipse.swt.widgets.TableColumn;
import org.eclipse.swt.widgets.Text;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IModelElement;
import org.fortiss.tooling.base.ui.annotation.AnnotationEntry;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ElementNameEditingSupport;
import org.fortiss.tooling.base.ui.annotation.valueprovider.IAnnotationValueProvider;
import org.fortiss.tooling.base.ui.widget.ExtendedCCombo;
import org.fortiss.tooling.kernel.model.INamedElement;

/**
 * <p>
 * This class provides a generic annotation view whose methods are intended to be reused by child
 * classes. Inherited classes can avoid code duplication, e.g. the creation of the table and its
 * first row, using this class.
 * </p>
 * <p>
 * In order to support displaying and editing specific annotation types, the corresponding
 * {@link IAnnotationValueProvider#createEditingSupport(org.eclipse.jface.viewers.ColumnViewer, Class, String)}
 * factory method must return an appropriate {@link EditingSupport}.
 * </p>
 * 
 * @author eder, diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating YELLOW Hash: D91763408246ECD3578DAD0E37BC4C41
 */
public class GenericAnnotationView extends AnnotationViewPartBase {
	/** Root composite of {@link GenericAnnotationView}. */
	Composite rootComposite;

	/** The table viewer */
	protected TableViewer tableViewer;

	/** Row an column filter for annotations */
	private AnnotationFilter annotationFilter;

	/** Widget for filter options */
	private AnnotationFilterWidget annotationFilterWidget;

	/**
	 * Last set of {@link AnnotationEntry}s used to {@link #update(Collection)} the
	 * {@link GenericAnnotationView}, i.e. during the construction of the, or because of a model
	 * change. The reference is required to update view after a change of the column (annotation)
	 * filter.
	 */
	Collection<AnnotationEntry> lastAnnotationEntries;

	/**
	 * Data required to identify a column displaying a particular {@link IAnnotatedSpecification} in
	 * a column of the {@link GenericAnnotationView}. Used to sort columns (see
	 * {@link #compareTo(ColumnHandle)}).
	 */
	class ColumnHandle implements Comparable<ColumnHandle> {
		/**
		 * Character used to separate the annotation name from the instance name for multi-instance
		 * annotations.
		 */
		private final static String MULTIINSTANCE_NAME_SEPARATOR = ":";

		/**
		 * {@link AnnotationEntry} from which a particular {@link IAnnotatedSpecification} is to be
		 * displayed in the column represented by this {@link ColumnHandle}.
		 */
		private AnnotationEntry entry;

		/**
		 * {@link IAnnotatedSpecification} from {@code entry} to be displayed in the column
		 * represented by this {@link ColumnHandle}.
		 */
		private IAnnotatedSpecification annotatedSpecification;

		/**
		 * Instance key identifying the particular instance of an {@link IAnnotatedSpecification}
		 * from {@code entry} to be displayed in the column represented by this {@link ColumnHandle}
		 * .
		 */
		private String instanceKey;

		/** Flag if this {@link ColumnHandle} represents a {@link CreateAnnotationInstanceColumn}. */
		private boolean isCreateColumn;

		/** Constructs a new {@link ColumnHandle}. */
		public ColumnHandle(AnnotationEntry entry, IAnnotatedSpecification annotatedSpecification,
				String instanceKey, boolean isCreateColumn) {

			this.entry = entry;
			this.annotatedSpecification = annotatedSpecification;
			this.instanceKey = instanceKey;
			this.isCreateColumn = isCreateColumn;
		}

		/**
		 * Returns the {@link AnnotationEntry} from which a particular
		 * {@link IAnnotatedSpecification} is to be
		 * displayed in the column represented by this {@link ColumnHandle}.
		 */
		public AnnotationEntry getEntry() {
			return entry;
		}

		/**
		 * Returns the {@link IAnnotatedSpecification} from {@code entry} to be displayed in the
		 * column represented by this {@link ColumnHandle}.
		 */
		public IAnnotatedSpecification getAnnotatedSpecification() {
			return annotatedSpecification;
		}

		/**
		 * Returns the instance key identifying the particular instance of an
		 * {@link IAnnotatedSpecification} from {@code entry} to be displayed in the column
		 * represented by this {@link ColumnHandle} .
		 */
		public String getInstanceKey() {
			return instanceKey;
		}

		/**
		 * Returns a flag if this {@link ColumnHandle} represents a
		 * {@link CreateAnnotationInstanceColumn}.
		 */
		public boolean isCreateColumn() {
			return isCreateColumn;
		}

		/**
		 * Returns the name of a column for a given {@link IAnnotatedSpecification} represented by
		 * the
		 * given {@link ColumnHandle}.
		 */
		protected String getColumnName() {
			// Determine column name
			String specName =
					getEntry().getSpecificationAnnotationName(
							getAnnotatedSpecification().getClass());

			if(getInstanceKey() != null) {
				if(!specName.isEmpty()) {
					specName += MULTIINSTANCE_NAME_SEPARATOR + " ";
				}
				specName += getInstanceKey();
			} else if(specName == null || specName.isEmpty()) {
				specName = "<Unnamed Annotation>";
			}
			return specName;
		}

		/** {@inheritDoc} */
		@Override
		public int compareTo(ColumnHandle other) {
			String columnName = getColumnName();
			String annotationName;

			int columnNameSeparatorIndex = columnName.indexOf(MULTIINSTANCE_NAME_SEPARATOR);
			if(columnNameSeparatorIndex != -1) {
				annotationName = columnName.substring(0, columnNameSeparatorIndex);
			} else {
				annotationName = columnName;
			}

			String otherColumnName = other.getColumnName();
			String otherAnnotationName;
			int otherColumnNameSeparatorIndex =
					otherColumnName.indexOf(MULTIINSTANCE_NAME_SEPARATOR);
			if(otherColumnNameSeparatorIndex != -1) {
				otherAnnotationName = otherColumnName.substring(0, otherColumnNameSeparatorIndex);
			} else {
				otherAnnotationName = otherColumnName;
			}

			// Place "create" columns (label: "<specification name>") after their instances
			// (which are labeled "<specification name>: <instance name>").
			if(annotationName.equals(otherAnnotationName)) {
				// Ensure uniqueness of "create" columns"
				if(isCreateColumn && other.isCreateColumn) {
					return 0;
				}
				if(isCreateColumn) {
					return 1;
				}
				if(other.isCreateColumn) {
					return -1;
				}
			}

			// Sort alphabetically by specification column label
			return columnName.compareTo(otherColumnName);
		}
	}

	/**
	 * Comparator that sorts rows according to hierarchy level. Elements within one hierarchy level
	 * are sorted by name.
	 */
	private class HierarchicalNameComparator extends ViewerComparator {

		/** {@inheritDoc} */
		@Override
		public int compare(Viewer viewer, Object o1, Object o2) {
			if(o1 instanceof AnnotationEntry && o2 instanceof AnnotationEntry) {
				IModelElement modelElement1 = ((AnnotationEntry)o1).getModelElement();
				IModelElement modelElement2 = ((AnnotationEntry)o2).getModelElement();
				int modelElement1Level = getModelElementLevel(modelElement1);
				int modelElement2Level = getModelElementLevel(modelElement2);

				EObject e1 = modelElement1;
				EObject e2 = modelElement2;

				if(modelElement1Level < modelElement2Level) {
					// Element1 is higher than element2: Get ancestor of element2 that is at the
					// level as the element1.
					EObject ancestor =
							getModelElementAncestor(modelElement2, modelElement2Level -
									modelElement1Level);

					if(modelElement1 == ancestor) {
						// Element1 is offspring of element2
						return -1;
					}

					// Further comparison starts on elements at the same level
					e2 = ancestor;
				} else if(modelElement1Level > modelElement2Level) {
					// Element2 is higher than element1: Get ancestor of element1 that is at the
					// level as the element2.
					EObject ancestor =
							getModelElementAncestor(modelElement1, modelElement1Level -
									modelElement2Level);

					if(modelElement2 == ancestor) {
						// Element2 is offspring of element1
						return 1;
					}

					// Further comparison starts on elements at the same level
					e1 = ancestor;
				}
				// Find direct children of common root element of element1 and element2...
				do {
					if(e1.eContainer() == e2.eContainer()) {
						break;
					}
					e1 = e1.eContainer();
					e2 = e2.eContainer();
				} while(true);

				// ... and compare them by name...
				if(e1 instanceof INamedElement && e2 instanceof INamedElement) {
					// Elements are at the same level -> compare by name
					return(((INamedElement)e1).getName().compareTo(((INamedElement)e2).getName()));
				}
			}

			// ... or hash code (as a fallback ordering)
			// int Integer.compare(int x, int y) requires JDK 1.7 which is not available
			// on the build server
			return new Integer(o1.hashCode()).compareTo(o2.hashCode());
		}
	}

	/** Widget to set content filters for the {@link GenericAnnotationView}. */
	private class AnnotationFilterWidget extends Composite {
		/** Annotation type filter: all annotations */
		private static final String ANNOTATION_TYPE_ANY = "Show all annotations";

		/** Combo box Column filter option based on annotation type */
		private ExtendedCCombo<Class<? extends IAnnotatedSpecification>> filterAnnotationTypeCombo;

		/** Constructs a new {@link AnnotationFilterWidget} */
		public AnnotationFilterWidget(Composite parent, int style, final GenericAnnotationView view) {
			super(parent, style);

			GridLayout layout = new GridLayout(4, false);
			layout.horizontalSpacing = 50;
			setLayout(layout);

			// Row 1: name filter scope / labels
			Composite nameFilterScopeComposite = createNameFilterScopeComposite(this, view);
			nameFilterScopeComposite.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, true,
					1, 1));

			final Label modelElementTypefilterLabel = new Label(this, SWT.NULL);
			modelElementTypefilterLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false,
					true, 1, 1));
			modelElementTypefilterLabel.setText("Filter model element type:");

			final Label hierarchyLevelFilterLabel = new Label(this, SWT.NULL);
			hierarchyLevelFilterLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true,
					1, 1));
			hierarchyLevelFilterLabel.setText("Filter model element hierarchy level:");

			final Label annotationTypeFilterLabel = new Label(this, SWT.NULL);
			annotationTypeFilterLabel.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true,
					1, 1));
			annotationTypeFilterLabel.setText("Filter annotation type:");

			// Row 2: controls
			Text nameFilterText = createNameFilterText(this, view);
			nameFilterText.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, true, 1, 1));

			Button modelElementTypeFilterButton =
					createModelElementTypeFilterOptionsCheckButton(this, view);
			modelElementTypeFilterButton.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false,
					true, 1, 1));

			CCombo hierarchyLevelFilterCombo =
					createHierarchyLevelFilterOptionsComboBox(this, view);
			hierarchyLevelFilterCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true,
					1, 1));

			ExtendedCCombo<Class<? extends IAnnotatedSpecification>> annotationTypeFilterCombo =
					createAnnotationTypeFilterOptionsComboBox(this, view);
			annotationTypeFilterCombo.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, false, true,
					1, 1));
		}

		/**
		 * Creates the {@link Text} input box for setting the pattern to filter the
		 * {@link GenericAnnotationView} for model element or annotation names
		 */
		private Text createNameFilterText(Composite parent, final GenericAnnotationView view) {
			// Text input field for filter pattern
			final Text nameFilterText =
					new Text(parent, SWT.SINGLE | SWT.BORDER | SWT.SEARCH | SWT.ICON_CANCEL);
			view.getAnnotationFilter();
			nameFilterText.setText(AnnotationFilter.FILTER_HINT_TEXT);
			nameFilterText.setForeground(Display.getCurrent().getSystemColor(SWT.COLOR_DARK_GRAY));

			// Observe changes of filter pattern
			nameFilterText.addModifyListener(new ModifyListener() {
				@Override
				public void modifyText(ModifyEvent e) {
					String filterPattern = nameFilterText.getText();
					if(filterPattern.isEmpty()) {
						nameFilterText.setText(AnnotationFilter.FILTER_HINT_TEXT);
						nameFilterText.selectAll();
						filterPattern = AnnotationFilter.FILTER_HINT_TEXT;
					}
					if(filterPattern.equals(AnnotationFilter.FILTER_HINT_TEXT)) {
						nameFilterText.setForeground(Display.getCurrent().getSystemColor(
								SWT.COLOR_DARK_GRAY));
					} else {
						nameFilterText.setForeground(Display.getCurrent().getSystemColor(
								SWT.COLOR_WIDGET_FOREGROUND));
					}
					view.getAnnotationFilter().setNameFilterPattern(filterPattern);
					view.update();
				}
			});

			// Select the filter pattern if it is equal to the FILTER_HINT_TEXT
			TextFocusListener textSelectionOnFocusListener = new TextFocusListener() {

				@Override
				protected void focusOut(Text text, Event e) {
					text.clearSelection();

				}

				@Override
				protected void focusIn(Text text, Event e) {
					if(text.getText().equals(AnnotationFilter.FILTER_HINT_TEXT) &&
							(e.type != SWT.MouseUp || text.getSelectionCount() == 0)) {
						text.selectAll();
					}

				}
			};
			nameFilterText.addListener(SWT.FocusIn, textSelectionOnFocusListener);
			nameFilterText.addListener(SWT.FocusOut, textSelectionOnFocusListener);
			nameFilterText.addListener(SWT.MouseDown, textSelectionOnFocusListener);
			nameFilterText.addListener(SWT.MouseUp, textSelectionOnFocusListener);

			return nameFilterText;
		}

		/**
		 * Creates a {@link Composite} that contains radio buttons to set the scope of the name
		 * filter (model element names or annotation names).
		 */
		private Composite createNameFilterScopeComposite(Composite parent,
				final GenericAnnotationView view) {
			Composite nameFilterScopeComposite = new Composite(parent, SWT.NULL);
			nameFilterScopeComposite.setLayout(new RowLayout());

			// Radio buttons to select between row and column name filtering
			final Label filterLabel = new Label(nameFilterScopeComposite, SWT.NULL);
			filterLabel.setText("Filter ");

			final Button filterComponentNamesButton =
					new Button(nameFilterScopeComposite, SWT.RADIO);
			filterComponentNamesButton.setText("model element");
			filterComponentNamesButton.setSelection(true);
			view.getAnnotationFilter().setFilterRowName(true);

			filterComponentNamesButton.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					// widgetSelected() is also fired when the selection is removed from the radio
					// button, so check, if this is the button that just obtained the selection
					if(!((Button)e.getSource()).getSelection()) {
						return;
					}
					view.getAnnotationFilter().setFilterRowName(true);
					view.getAnnotationFilter().setFilterColumnName(false);
					view.update();
				}
			});

			final Button filterAnnotationNamesButton =
					new Button(nameFilterScopeComposite, SWT.RADIO);
			filterAnnotationNamesButton.setText("annotation names:");
			filterAnnotationNamesButton.setSelection(false);

			filterAnnotationNamesButton.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					// widgetSelected() is also fired when the selection is removed from the radio
					// button, so check, if this is the button that just obtained the selection
					if(!((Button)e.getSource()).getSelection()) {
						return;
					}
					view.getAnnotationFilter().setFilterRowName(false);
					view.getAnnotationFilter().setFilterColumnName(true);
					view.update();
				}
			});

			return nameFilterScopeComposite;
		}

		/**
		 * Returns a {@link Button} for setting the options for filtering the
		 * {@link GenericAnnotationView} for the currently selected model element type.
		 * 
		 * @param view
		 *            {@link GenericAnnotationView} to be filtered.
		 */
		private Button createModelElementTypeFilterOptionsCheckButton(Composite parent,
				final GenericAnnotationView view) {

			// Check box to restrict view to currently selected model element type
			final Button filterModelElementTypeButton = new Button(parent, SWT.CHECK);

			filterModelElementTypeButton.setText("Show only selected model element type.");
			filterModelElementTypeButton.setSelection(false);

			filterModelElementTypeButton.addSelectionListener(new SelectionAdapter() {
				@Override
				public void widgetSelected(SelectionEvent e) {
					// widgetSelected() is also fired when the selection is removed from the radio
					// button, so check, if this is the button that just obtained the selection
					view.getAnnotationFilter().setRestrictToSelectedModelElementType(
							((Button)e.getSource()).getSelection());
					view.update();
				}
			});

			return filterModelElementTypeButton;
		}

		/**
		 * Returns a {@link CCombo} for setting the options for filtering the
		 * {@link GenericAnnotationView} for the model element hierarchy level.
		 */
		private CCombo createHierarchyLevelFilterOptionsComboBox(Composite parent,
				final GenericAnnotationView view) {

			// Combo box to select component hierarchy level filter
			final CCombo filterHierarchyLevelCombo = new CCombo(parent, SWT.READ_ONLY);

			filterHierarchyLevelCombo.add(AnnotationFilter.HIERARCHY_LEVELS_ALL);
			filterHierarchyLevelCombo.add(AnnotationFilter.HIERARCHY_LEVELS_CURRENT);
			filterHierarchyLevelCombo.add(AnnotationFilter.HIERARCHY_LEVELS_SELECTED_SUBMODEL);
			filterHierarchyLevelCombo.select(0);

			view.getAnnotationFilter().setHierarchyLevelFilter(filterHierarchyLevelCombo.getText());

			filterHierarchyLevelCombo.addSelectionListener(new SelectionAdapter() {
				/** {@inheritDoc} */
				@Override
				public void widgetSelected(SelectionEvent e) {
					view.getAnnotationFilter().setHierarchyLevelFilter(
							filterHierarchyLevelCombo.getText());
					view.update();
				}
			});

			return filterHierarchyLevelCombo;
		}

		/**
		 * Returns an {@link ExtendedCCombo} for setting the options for filtering the
		 * {@link GenericAnnotationView} for the annotation type.
		 */
		private ExtendedCCombo<Class<? extends IAnnotatedSpecification>>
				createAnnotationTypeFilterOptionsComboBox(Composite parent,
						final GenericAnnotationView view) {

			// // Combo box to select annotation type filter
			filterAnnotationTypeCombo =
					new ExtendedCCombo<Class<? extends IAnnotatedSpecification>>(parent,
							SWT.READ_ONLY);
			updateAnnotationTypeFilterOptionsComboBox(lastAnnotationEntries, view);

			filterAnnotationTypeCombo.addSelectionListener(new SelectionAdapter() {
				/** {@inheritDoc} */
				@Override
				public void widgetSelected(SelectionEvent e) {
					view.getAnnotationFilter().setAnnotationTypeFilter(
							filterAnnotationTypeCombo.getObject(filterAnnotationTypeCombo
									.getSelectionIndex()));
					view.update();
				}
			});

			return filterAnnotationTypeCombo;
		}

		/**
		 * Update the combobox that can be used to configure the annotation type filter based on the
		 * annotations contained in the current set of annotation entries / model elements (before
		 * filtering).
		 */
		public void updateAnnotationTypeFilterOptionsComboBox(
				Collection<AnnotationEntry> annotationEntries, GenericAnnotationView view) {
			String currentSelection = filterAnnotationTypeCombo.getText();
			filterAnnotationTypeCombo.removeAll();
			filterAnnotationTypeCombo.add(ANNOTATION_TYPE_ANY, null);

			if(annotationEntries != null) {
				// Set needed to ensure that each IAnnotatedSpecification which may be contributed
				// by more than one AnnotationEntry is added exactly once to the ComboBox.
				Set<Class<? extends IAnnotatedSpecification>> annotatedSpecificationClasses =
						new HashSet<Class<? extends IAnnotatedSpecification>>();

				for(AnnotationEntry annotationEntry : annotationEntries) {
					for(IAnnotatedSpecification spec : pickInstanceOf(
							IAnnotatedSpecification.class, annotationEntry.getSpecificationsList())) {
						if(!annotatedSpecificationClasses.contains(spec.getClass())) {
							filterAnnotationTypeCombo.add(
									"Show " +
											annotationEntry.getSpecificationAnnotationName(spec
													.getClass()), spec.getClass());
							annotatedSpecificationClasses.add(spec.getClass());
						}
					}
				}
			}

			// If possible, maintain selected element
			int index = filterAnnotationTypeCombo.indexOf(currentSelection);
			if(index == -1) {
				filterAnnotationTypeCombo.select(0);
				view.getAnnotationFilter().setAnnotationTypeFilter(
						filterAnnotationTypeCombo.getObject(0));
			} else {
				filterAnnotationTypeCombo.select(index);
			}

			// Adjust size to new content
			filterAnnotationTypeCombo.pack(true);
		}
	}

	/**
	 * Sets a fixed {@link AnnotationFilter}.
	 * <ul>
	 * <li>If the {@link AnnotationFilter} is non-{@code null}, it sets the filter and disables the
	 * {@link AnnotationFilterWidget}. For instance, this can be used to embed / extend the
	 * {@link GenericAnnotationView} into a specialized view dedicated to a particular model type.</li>
	 * <li>Otherwise, this method reinitializes the {@link AnnotationFilterWidget}.</li>
	 * </ul>
	 */
	public void setFixedAnnotationFilter(AnnotationFilter annotationFilter) {

		// Remove old table row filter
		tableViewer.removeFilter(this.annotationFilter);

		if(annotationFilter != null) {
			// Externally created filter -> remove filter option GUI
			this.annotationFilter = annotationFilter;
			if(annotationFilterWidget != null) {
				annotationFilterWidget.dispose();
				annotationFilterWidget = null;
			}
		} else {
			// No externally created filter -> setup filter option GUI
			this.annotationFilter = new AnnotationFilter();
			setupAnnotationFilterWidget();
		}

		// Add new table row filter
		tableViewer.addFilter(this.annotationFilter);

		// Update layout
		rootComposite.layout(true, true);
	}

	/** Returns the view's {@link AnnotationFilter}. */
	private AnnotationFilter getAnnotationFilter() {
		return annotationFilter;
	}

	/** Computes the level of an {@link IModelElement} in the tree */
	public static int getModelElementLevel(EObject o) {
		int level = 0;
		while(o.eContainer() != null) {
			level++;
			o = o.eContainer();
		}

		return level;
	}

	/**
	 * Returns the ancestor of the {@link IModelElement} {@code modelElement} that is
	 * {@code levelsUp} levels above model element in the hierarchy, or {@code null}, in case
	 * {@code levelsUp} > depth(modelElement}.
	 */
	public static EObject getModelElementAncestor(EObject o, int levelsUp) {
		while(o.eContainer() != null && levelsUp > 0) {
			o = o.eContainer();
			levelsUp--;
		}
		return o;

	}

	/** Updates the table after a filter change */
	private void update() {
		// Full update, i.e. including reconstructions of columns is always required due to the
		// following scenario:
		// - Filter is set for annotation names.
		// - Filter scope is changed to component names -> annotation filter needs to be undone
		if(lastAnnotationEntries != null) {
			update(lastAnnotationEntries);
		}
	}

	/** {@inheritDoc} */
	@Override
	protected void update(Collection<AnnotationEntry> annotationEntries) {
		// Update filter
		annotationFilter.setCurrentySelectedModelElement(getCurrentlySelectedObject());
		if(annotationFilterWidget != null) {
			annotationFilterWidget.updateAnnotationTypeFilterOptionsComboBox(annotationEntries,
					this);
		}

		// Remember last annotation entries (required to re-draw table after changing the
		// column/annotation filter
		lastAnnotationEntries = annotationEntries;

		// Minimize flickering while updating the table
		tableViewer.getTable().setRedraw(false);

		// Input must be set before the construction of the columns, since the editing support for a
		// given cell might need to query the contents of other model elements (e.g., in order to
		// aggregate a set of admissible choices for its value).
		tableViewer.setInput(annotationEntries);

		/*
		 * Sorted set of {@link ColumnHandle} used to instantiate this {@link
		 * GenericAnnotationView}'s columns in the right order.
		 */
		Set<ColumnHandle> sortedColumnHandles = new TreeSet<ColumnHandle>();

		/*
		 * Dispose all columns of the table viewer, except the first column, which holds the names
		 * of the model elements
		 */
		while(tableViewer.getTable().getColumnCount() > 1) {
			tableViewer.getTable().getColumn(1).dispose();
		}

		// Aggregate required columns. Column order is defined by ColumnHandle.compareTo().
		for(AnnotationEntry entry : annotationEntries) {
			for(IAnnotatedSpecification spec : entry.getSpecificationsList()) {
				// Create new column for each instance of the current multi-instance annotation
				if((entry.allowsDynamicAnnotationInstances(spec.getClass())) ||
						!entry.getInstanceKeys(spec.getClass()).isEmpty()) {

					for(String instanceKey : entry.getInstanceKeys(spec.getClass())) {
						ColumnHandle columnHandle =
								new ColumnHandle(entry, spec, instanceKey, false);
						if(annotationFilter.passesColumnFilter(columnHandle)) {
							sortedColumnHandles.add(columnHandle);
						}
					}
				}

				// Create a new column for
				// - ordinary (single-instance) annotations
				// - "create" column for multi-instance annotations for which instances can be
				// created dynamically
				if(entry.getInstanceKeys(spec.getClass()).isEmpty() ||
						entry.allowsDynamicAnnotationInstances(spec.getClass())) {
					ColumnHandle columnHandle =
							new ColumnHandle(entry, spec, null,
									entry.allowsDynamicAnnotationInstances(spec.getClass()));

					if(annotationFilter.passesColumnFilter(columnHandle)) {
						sortedColumnHandles.add(columnHandle);
					}

				}
			}
		}

		// Instantiate columns
		for(ColumnHandle columnHandle : sortedColumnHandles) {
			createAnnotationColumn(columnHandle);
		}

		// Update the view
		tableViewer.getTable().setRedraw(true);
		tableViewer.refresh();
	}

	/** Initializes the {@link AnnotationFilterWidget}. */
	private void setupAnnotationFilterWidget() {
		annotationFilterWidget = new AnnotationFilterWidget(rootComposite, SWT.NULL, this);
		annotationFilterWidget.setLayoutData(new GridData(SWT.FILL, SWT.CENTER, true, false, 1, 1));
	}

	/** {@inheritDoc} */
	@Override
	public void createPartControl(Composite parent) {
		super.createPartControl(parent);
		rootComposite = new Composite(parent, SWT.NULL);
		rootComposite.setLayout(new GridLayout(1, false));

		tableViewer = new TableViewer(rootComposite, SWT.BORDER | SWT.FULL_SELECTION);
		Table table = tableViewer.getTable();
		table.setHeaderVisible(true);
		table.setLinesVisible(true);
		table.setLayoutData(new GridData(SWT.FILL, SWT.FILL, true, true, 1, 1));

		// Install layout for Table in order to ensure that it claims all available vertical space
		// in case the
		// filter option GUI is removed (see setFixedAnnotationFilter()).
		table.setLayout(new FillLayout(SWT.VERTICAL));

		createModelElementColumn();

		tableViewer.setComparator(new HierarchicalNameComparator());

		// Rows (= components) are filtered using a specialized ViewerFilter.
		// Since the data model is an array of AnnotationEntries, columns (= annotations)
		// cannot be filtered using such a filter, and require the reconstruction of the
		// table, with an implementation of the filter in update(Set<AnnotationEntry>).
		annotationFilter = new AnnotationFilter();
		tableViewer.addFilter(annotationFilter);

		// Add filter widget
		setupAnnotationFilterWidget();
	}

	/** Creates the (leading) column which displays the model elements */
	protected void createModelElementColumn() {

		final TableViewerColumn firstColumn = new TableViewerColumn(tableViewer, SWT.NONE);
		firstColumn.getColumn().setText("Model Element");
		firstColumn.getColumn().setWidth(100);

		tableViewer.setContentProvider(new ArrayContentProvider());
		firstColumn.setLabelProvider(new ElementNameLabelProvider(this));
		firstColumn.setEditingSupport(new ElementNameEditingSupport(tableViewer));
	}

	/** Creates a column in tableViewer that displays the annotation given in spec */
	protected void createAnnotationColumn(ColumnHandle columnHandle) {

		TableColumn tableColumn = null;
		if(columnHandle.isCreateColumn()) {
			// Create column that contains a button for adding new instances of the
			// current annotation type.

			CreateAnnotationInstanceColumn column =
					new CreateAnnotationInstanceColumn(tableViewer, SWT.NONE,
							columnHandle.getAnnotatedSpecification(), columnHandle.getEntry());
			tableColumn = column.getColumn();

		} else {
			Class<? extends IAnnotatedSpecification> annotationClass =
					columnHandle.getAnnotatedSpecification().getClass();

			// Add new column
			try {

				// Have the matching EditingSupport created for the current annotated specification
				// (work is delegated to respective IAnnotationValueProvider implementation)
				EditingSupport editingSupport =
						columnHandle.getEntry().createSpecificationEditElement(tableViewer,
								annotationClass, columnHandle.getInstanceKey());
				TableViewerColumn column = new TableViewerColumn(tableViewer, SWT.NONE);
				tableColumn = column.getColumn();

				String specName = columnHandle.getColumnName();
				tableColumn.setText(specName);

				column.setEditingSupport(editingSupport);

				// Add column label provider, and set EditingSupport
				column.setLabelProvider(new AnnotationLabelProvider(annotationClass, this,
						columnHandle.getInstanceKey(), editingSupport));
			} catch(Exception e) {
				System.out.println("Failed to create editing support for annotation class " +
						annotationClass.getCanonicalName() + ": " + e.getMessage());
			}

		}

		if(tableColumn != null) {
			tableColumn.setWidth(125);
		}

	}

	/** {@inheritDoc} */
	@Override
	public void setFocus() {
		// nothing to do
	}

	/** {@inheritDoc} */
	@Override
	public void dispose() {
		tableViewer.getTable().dispose();
		super.dispose();
	}

}
