/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2014 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.widget;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.swt.custom.CCombo;
import org.eclipse.swt.widgets.Composite;

/**
 * Extension of {@link CCombo} that links the combo box entries with foreign data that can be
 * added using {@link #add(String, Object)} and queried using {@link #getObject(int)}.
 * 
 * @author diewald, barner
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating RED Hash:
 */
public class ExtendedCCombo<T> extends CCombo {

	/** Correlates the entries in the {@link CCombo} with arbitrary objects */
	private List<T> objectReferenceList;

	/** Constructs an {@link ExtendedCCombo} */
	public ExtendedCCombo(Composite parent, int style) {
		super(parent, style);
		objectReferenceList = new ArrayList<T>();
	}

	/**
	 * Adds a {@code comboString} to the {@link CCombo} and references the object
	 * {@code referencedObject} for this entry.
	 */
	public void add(String comboString, T referencedObject) {
		super.add(comboString);

		objectReferenceList.add(referencedObject);
	}

	/**
	 * <p>
	 * {@inheritDoc}
	 * </p>
	 * <p>
	 * <b>NOTE:</b>This method references {@code null} with the given {@code comboString}. Use
	 * {@link #add(String, Object)} for actually linking an object with a given {@code comboString}.
	 * </p>
	 */
	@Override
	public void add(String comboString, int index) {
		super.add(comboString, index);

		objectReferenceList.add(index, null);
	}

	/**
	 * <p>
	 * {@inheritDoc}
	 * </p>
	 * <p>
	 * <b>NOTE:</b>This method references {@code null} with the given {@code comboString}. Use
	 * {@link #add(String, Object)} for actually linking an object with a given {@code comboString}.
	 * </p>
	 */
	@Override
	public void add(String string) {
		add(string, null);
	}

	/** {@inheritDoc} */
	@Override
	public void removeAll() {
		super.removeAll();
		objectReferenceList.clear();
	}

	/** {@inheritDoc} */
	@Override
	public void remove(int index) {
		super.remove(index);
		objectReferenceList.remove(index);
	}

	/** {@inheritDoc} */
	@Override
	public void remove(int start, int end) {
		super.remove(start, end);
		// Elements are shifted to the left
		// i <= end: super.remove() works in the range [start; end].
		for(int i = start; i <= end; i++) {
			objectReferenceList.remove(start);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void remove(String string) {
		int index = indexOf(string);
		if(index != -1) {
			super.remove(string);
			objectReferenceList.remove(index);
		}
	}

	/** Returns the object (type T) stored at the given index. */
	public T getObject(int index) {
		return objectReferenceList.get(index);
	}

	/** {@inheritDoc} */
	@Override
	protected void checkSubclass() {
		// Allow subclassing by overriding this check.
	}
}
