/*--------------------------------------------------------------------------+
$Id$
|                                                                          |
| Copyright 2012 ForTISS GmbH                     |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.graphicsGL.util;

import static org.fortiss.tooling.graphicsGL.util.Graphics3DModelElementFactory.vector;
import static org.fortiss.tooling.kernel.utils.EcoreUtils.copy;

import java.util.List;

import org.fortiss.tooling.graphicsGL.model.Color3D;
import org.fortiss.tooling.graphicsGL.model.mesh.Mesh3D;
import org.fortiss.tooling.graphicsGL.model.scene.Camera3D;
import org.fortiss.tooling.graphicsGL.model.scene.Light3D;
import org.fortiss.tooling.graphicsGL.model.scene.Object3D;
import org.fortiss.tooling.graphicsGL.model.scene.Scene3D;
import org.fortiss.tooling.graphicsGL.model.scene.SceneFactory;
import org.fortiss.tooling.graphicsGL.model.scene.ScenePackage;

/**
 * Model element factory for {@link ScenePackage}.
 * 
 * @author hoelzl
 * @author $Author$
 * @version $Rev$
 * @ConQAT.Rating GREEN Hash: 998BB1F67732795A2D79987FAE3D8258
 */
public class Scene3DModelElementFactory {

	/** Creates a 3D object with a mesh and material. */
	public static Object3D object3D(float x, float y, float z, Mesh3D mesh) {
		Object3D o = SceneFactory.eINSTANCE.createObject3D();
		o.setMesh(mesh);
		o.setPosition(vector(x, y, z));
		return o;
	}

	/** Creates a named 3D object with a mesh and material. */
	public static Object3D object3D(String name, String comment, float x, float y, float z,
			Mesh3D mesh) {
		Object3D o = object3D(x, y, z, mesh);
		o.setName(name);
		o.setComment(comment);
		return o;
	}

	/** Creates a 3D light at the given position and the given diffuse color. */
	public static Light3D light3D(float x, float y, float z, Color3D color) {
		Light3D l = SceneFactory.eINSTANCE.createLight3D();
		l.setPosition(vector(x, y, z));
		l.setDiffuse(copy(color));
		return l;
	}

	/** Creates a 3D light at the given position and the given light component colors. */
	public static Light3D light3D(float x, float y, float z, Color3D ambient, Color3D diffuse,
			Color3D specular) {
		Light3D l = SceneFactory.eINSTANCE.createLight3D();
		l.setPosition(vector(x, y, z));
		l.setAmbient(copy(ambient));
		l.setDiffuse(copy(diffuse));
		l.setSpecular(copy(specular));
		l.setAttenuation(vector(1.0f, 0.0f, 0.0f));
		return l;
	}

	/** Creates a named 3D light at the given position and the given diffuse color. */
	public static Light3D light3D(String name, String comment, float x, float y, float z,
			Color3D ambient, Color3D diffuse, Color3D specular) {
		Light3D l = light3D(x, y, z, ambient, diffuse, specular);
		l.setName(name);
		l.setComment(comment);
		return l;
	}

	/** Creates a camera at the given position looking at the given point. */
	public static Camera3D
			camera3D(float posX, float posY, float posZ, float tx, float ty, float tz) {
		Camera3D cam = SceneFactory.eINSTANCE.createCamera3D();
		cam.setPosition(vector(posX, posY, posZ));
		cam.setLookAtLocation(vector(tx, ty, tz));
		return cam;
	}

	/** Creates a named camera with the given data. */
	public static Camera3D camera3d(String name, String comment, float posX, float posY,
			float posZ, float tx, float ty, float tz) {
		Camera3D c = camera3D(posX, posY, posZ, tx, ty, tz);
		c.setName(name);
		c.setComment(comment);
		return c;
	}

	/** Creates an empty scene. */
	public static Scene3D scene3D() {
		return SceneFactory.eINSTANCE.createScene3D();
	}

	/** Creates a scene with the given objects, lights, and cameras. */
	public static Scene3D scene3D(List<Object3D> objects, List<Light3D> lights,
			List<Camera3D> cameras) {
		Scene3D scene = scene3D();
		scene.getObjects().addAll(objects);
		scene.getLights().addAll(lights);
		scene.getCameras().addAll(cameras);
		return scene;
	}
}
